# coding=utf-8
import datetime
import locale
import platform
from contextlib import contextmanager

from django.db import models

from .constants import MAIL_TEMPLATE_CONTEXT_CHOICES, MAIL_TEMPLATE_CONTEXT_APPOINTMENT, \
    MAIL_TEMPLATE_CONTEXT_SUBJECT, MAIL_TEMPLATE_CONTEXT_VISIT, MAIL_TEMPLATE_CONTEXT_VOUCHER, DEFAULT_LOCALE_NAME
from ..docx_helper import process_file
from ..models import Appointment, Visit, StudySubject, Worker, Voucher

DATE_FORMAT_FULL = "%A %d %B %Y"

DATETIME_FORMAT = "%A %d %B %Y, %H:%m"

DATE_FORMAT_SHORT = "%d.%m.%Y"

DATE_FORMAT_TIME = "%H:%M"

now = datetime.datetime.now()


def date_format_encoding():
    return locale.getlocale(locale.LC_TIME)[1] or locale.getpreferredencoding()


@contextmanager
def setlocale(name):
    saved = locale.setlocale(locale.LC_TIME)
    try:
        yield locale.setlocale(locale.LC_TIME, name)
    finally:
        locale.setlocale(locale.LC_TIME, saved)


def get_formatted_time(time_format):
    # TODO: 2TO3 removed string decoding using the locale
    # Was:
    # return now.strftime(time_format).decode(date_format_encoding())
    # Is:
    return now.strftime(time_format)  # /-->


def date_to_str(date, format):
    if date is not None:
        # TODO: 2TO3 removed string decoding using the locale
        # Was:
        # return date.strftime(format).decode(date_format_encoding())
        # Is:
        return date.strftime(format)  # /-->
    else:
        return ""


class MailTemplate(models.Model):
    MAILS_TEMPLATE_GENERIC_TAGS = [
        ("##DATE_FULL##", "Current date when the mail will be generated (long format)",
         get_formatted_time(DATE_FORMAT_FULL)),
        ("##DATE_SHORT##", "Current date when the mail will be generated (short format)",
         now.strftime(DATE_FORMAT_SHORT)),
        ("##WORKER##", "The full name of the currently logged in user", ""),
        ("##WORKER_EMAIL##", "Email address of the currently logged in user", "")
    ]

    MAILS_TEMPLATE_SUBJECT_TAGS = [
        ("##S_FULL_NAME##", "Subject's full name", "first_name last_name"),
        ("##S_FIRST_NAME##", "Subject's first name", ""),
        ("##S_LAST_NAME##", "Subject's last name", ""),
        ("##S_ADDRESS##", "Subject's address", "street name and number"),
        ("##S_CITY##", "Subject's city of residence", ""),
        ("##S_POST_CODE##", "Subject's post code of residence", ""),
        ("##S_COUNTRY##", "Subject's country of residence", ""),
        ("##S_SEX##", "Subject's gender", "Male/Female"),
        ("##S_TYPE##", "Subject's type", "CONTROL/PATIENT"),
        ("##S_DATE_BORN##", "Subject's date of birth", get_formatted_time(DATE_FORMAT_SHORT)),

        ("##S_EMAIL##", "Subject's email address", ""),
        ("##S_PHONE_NUMBER##", "Subject's phone number", ""),
        ("##S_PHONE_NUMBER_2##", "Subject's second phone number", ""),
        ("##S_PHONE_NUMBER_3##", "Subject's third phone number", ""),
        ("##S_MAIL_LANGUAGE##", "Subject's preferred language for written communication", ""),
        ("##S_KNOWN_LANGUAGES##", "List of languages known by the subject", "comma separated"),

        ("##S_SCREENING_NUMBER##", "Subject's screening number", ""),
        ("##S_DIAGNOSIS##", "Subject's diagnosis", ""),
        ("##S_DIAGNOSIS_YEAR##", "Subject's year of diagnosis", ""),
        ("##S_MPOWER_ID##", "Subject's mPower identifier", ""),
        ("##S_ND_NUMBER##", "Subject's ND number", ""),
        ("##S_DATE_ADDED##", "Subject's date of creation", get_formatted_time(DATE_FORMAT_SHORT)),

        ("##S_VIRUS_1_SAMPLE_COLLECTION_DATE##", "Virus test collection date in visit 1",
         get_formatted_time(DATE_FORMAT_SHORT)),
        ("##S_VIRUS_2_SAMPLE_COLLECTION_DATE##", "Virus test collection date in visit 2",
         get_formatted_time(DATE_FORMAT_SHORT)),
        ("##S_VIRUS_3_SAMPLE_COLLECTION_DATE##", "Virus test collection date in visit 3",
         get_formatted_time(DATE_FORMAT_SHORT)),
        ("##S_VIRUS_4_SAMPLE_COLLECTION_DATE##", "Virus test collection date in visit 4",
         get_formatted_time(DATE_FORMAT_SHORT)),
        ("##S_VIRUS_5_SAMPLE_COLLECTION_DATE##", "Virus test collection date in visit 5",
         get_formatted_time(DATE_FORMAT_SHORT)),

        ("##S_VIRUS_1_RESULT##", "Virus test results in visit 1", "Positive"),
        ("##S_VIRUS_2_RESULT##", "Virus test results in visit 2", "Negative"),
        ("##S_VIRUS_3_RESULT##", "Virus test results in visit 3", "Inconclusive"),
        ("##S_VIRUS_4_RESULT##", "Virus test results in visit 4", ""),
        ("##S_VIRUS_5_RESULT##", "Virus test results in visit 5", ""),

        ("##S_VIRUS_1_IGA_STATUS##", "IgA Status in visit 1", "Positive"),
        ("##S_VIRUS_2_IGA_STATUS##", "IgA Status in visit 2", "Negative"),
        ("##S_VIRUS_3_IGA_STATUS##", "IgA Status in visit 3", "Borderline"),
        ("##S_VIRUS_4_IGA_STATUS##", "IgA Status in visit 4", ""),
        ("##S_VIRUS_5_IGA_STATUS##", "IgA Status in visit 5", ""),

        ("##S_VIRUS_1_IGG_STATUS##", "IgG Status in visit 1", "Positive"),
        ("##S_VIRUS_2_IGG_STATUS##", "IgG Status in visit 2", "Negative"),
        ("##S_VIRUS_3_IGG_STATUS##", "IgG Status in visit 3", "Borderline"),
        ("##S_VIRUS_4_IGG_STATUS##", "IgG Status in visit 4", ""),
        ("##S_VIRUS_5_IGG_STATUS##", "IgG Status in visit 5", ""),

        ("##S_HEALTH_PARTNER_NAME##", "Name of the health partner", ""),
        ("##S_HEALTH_PARTNER_ADDRESS##", "Address of the health partner", ""),
        ("##S_HEALTH_PARTNER_ZIP_CODE##", "Zip code of the health partner", ""),
        ("##S_HEALTH_PARTNER_CITY##", "City of the health partner", ""),

    ]

    MAILS_TEMPLATE_VISIT_TAGS = [
        ("##V_DATE_START_FULL##", "Visit's start date", get_formatted_time(DATETIME_FORMAT)),
        ("##V_DATE_START_SHORT##", "Visit's start date", get_formatted_time(DATE_FORMAT_SHORT)),
        ("##V_DATE_ENDS_FULL##", "Visit's end date", get_formatted_time(DATETIME_FORMAT)),
        ("##V_DATE_ENDS_SHORT##", "Visit's end date", get_formatted_time(DATE_FORMAT_SHORT)),
    ]

    MAILS_TEMPLATE_APPOINTMENT_TAGS = [
        ("##A_DATE_FULL##", "Appointment's date and time", get_formatted_time(DATETIME_FORMAT)),
        ("##A_DATE_SHORT##", "Appointment's date", get_formatted_time(DATE_FORMAT_SHORT)),
        ("##A_TIME##", "Appointment's time", get_formatted_time(DATE_FORMAT_TIME)),
        ("##A_FLYING_TEAM##", "Appointment's flying team location", ""),
        ("##A_LOCATION##", "Appointment's location", "value can be 'Flying Team'"),
        ("##A_LOCATION_OR_FLYINGTEAM##", "Appointment's real location",
         "if flying team then returns flying team exact location, otherwise returns location name"),
        ("##A_STATUS##", "Appointment's status", ""),
        ("##A_WORKER##", "Worker conducting the assessment", "first_name last_name"),
        ("##A_WORKER_PHONE##", "Phone number of the worker conducting the assessment", ""),
        ("##A_WORKER_EMAIL##", "Email address of the worker conducting the assessment", ""),
        ("##A_ROOM##", "Appointment's room", 'room_number address city'),
        ("##A_LENGTH##", "Appointment's duration", 'integer, value in minutes'),
        ("##A_TYPES##", "Appointment's types", "comma separated"),
    ]

    MAILS_TEMPLATE_VOUCHER_TAGS = [
        ("##C_NUMBER##", "Number", ''),
        ("##C_PATIENT_NAME##", "Voucher Partner name", ''),
        ("##C_VOUCHER_TYPE##", "Voucher type", ''),

        ("##C_ISSUE_DATE_SHORT##", "Issue date", get_formatted_time(DATE_FORMAT_SHORT)),
        ("##C_EXPIRY_START_SHORT##", "Expiry date", get_formatted_time(DATE_FORMAT_SHORT)),

        ("##C_PARTNER_NAME##", "Voucher Partner name", ''),
        ("##C_PARTNER_ADDRESS##", "Voucher Partner address", ''),
        ("##C_PARTNER_CITY##", "Voucher Partner city", ''),
        ("##C_PARTNER_POSTAL_CODE##", "Voucher Partner postal code", ''),
        ("##C_PARTNER_COUNTRY##", "Voucher Partner country", ''),
        ("##C_PARTNER_PHONE##", "Voucher Partner phone", ''),
        ("##C_HOURS##", "Hours", ''),
    ]

    name = models.CharField(max_length=255)
    context = models.CharField(max_length=1, choices=MAIL_TEMPLATE_CONTEXT_CHOICES)
    language = models.ForeignKey("web.Language", on_delete=models.CASCADE, blank=True, null=True)
    template_file = models.FileField(upload_to='templates/')

    @staticmethod
    def get_appointment_mail_templates(languages):
        return MailTemplate.get_mail_templates_for_context(languages, MAIL_TEMPLATE_CONTEXT_APPOINTMENT)

    @staticmethod
    def get_subject_mail_templates(languages):
        return MailTemplate.get_mail_templates_for_context(languages, MAIL_TEMPLATE_CONTEXT_SUBJECT)

    @staticmethod
    def get_voucher_mail_templates(languages):
        return MailTemplate.get_mail_templates_for_context(languages, MAIL_TEMPLATE_CONTEXT_VOUCHER)

    @staticmethod
    def get_visit_mail_templates(languages):
        return MailTemplate.get_mail_templates_for_context(languages, MAIL_TEMPLATE_CONTEXT_VISIT)

    @staticmethod
    def get_mail_templates_for_context(languages, context):
        languages_names = [language.name for language in languages]
        templates = list(MailTemplate.objects.filter(context=context).all())
        active_templates = []
        disabled_templates = []
        for template in templates:
            if template.language is None:
                if len(languages) == 0:
                    active_templates.append(template)
                else:
                    disabled_templates.append(template)
            elif template.language.name in languages_names:
                active_templates.append(template)
            else:
                disabled_templates.append(template)
        active_templates.sort(key=lambda x: languages_names.index(x.language.name) if x.language is not None else -1)
        return active_templates, disabled_templates

    def apply(self, instance, user, stream):
        appointment = None
        visit = None
        study_subject = None
        voucher = None
        if isinstance(instance, Appointment):
            appointment = instance
            visit = instance.visit
            study_subject = visit.subject
        elif isinstance(instance, Visit):
            visit = instance
            study_subject = visit.subject
        elif isinstance(instance, StudySubject):
            study_subject = instance
        elif isinstance(instance, Voucher):
            voucher = instance
        # set locale to get correct date format
        locale_name = self.get_locale_name()
        with setlocale(locale_name):
            replacements = {}
            replacements.update(self.get_generic_replacements(Worker.get_by_user(user)))
            replacements.update(self.get_appointment_replacements(appointment))
            replacements.update(self.get_visit_replacements(visit))
            replacements.update(self.get_subject_replacements(study_subject))
            replacements.update(self.get_voucher_replacements(voucher))
            process_file(self.template_file.path, stream, replacements)
        return stream

    def get_locale_name(self):
        if self.language is None:
            locale_name = DEFAULT_LOCALE_NAME
        else:
            locale_name = self.language.locale
            if platform.system() == 'Windows':
                locale_name = self.language.windows_locale_name
        return locale_name

    @staticmethod
    def get_generic_replacements(worker):
        current_datetime = datetime.datetime.now()
        email = ''
        if worker is not None:
            email = worker.email
        return {
            # TODO: 2to3
            # Was:
            # "##DATE_FULL##": current_datetime.strftime(DATE_FORMAT_FULL).decode(date_format_encoding()),
            # "##DATE_SHORT##": current_datetime.strftime(DATE_FORMAT_SHORT).decode(date_format_encoding()),
            # Is:
            "##DATE_FULL##": current_datetime.strftime(DATE_FORMAT_FULL),
            "##DATE_SHORT##": current_datetime.strftime(DATE_FORMAT_SHORT),
            # /-->
            "##WORKER##": str(worker),
            "##WORKER_EMAIL##": email
        }

    @staticmethod
    def get_appointment_replacements(appointment):
        if appointment is None:
            return {}

        if appointment.worker_assigned is not None:
            worker_phone_number = appointment.worker_assigned.phone_number
            worker_email_address = appointment.worker_assigned.email
        else:
            worker_phone_number = ""
            worker_email_address = ""
        if appointment.datetime_when is not None:
            # TODO: 2to3
            # Was:
            # appointment_date_full = appointment.datetime_when.strftime(DATETIME_FORMAT).decode(
            #    date_format_encoding())
            # appointment_date_short = appointment.datetime_when.strftime(DATE_FORMAT_SHORT).decode(
            #    date_format_encoding())
            # appointment_date_time = appointment.datetime_when.strftime(DATE_FORMAT_TIME).decode(
            #    date_format_encoding())
            # Is:
            appointment_date_full = appointment.datetime_when.strftime(DATETIME_FORMAT)
            appointment_date_short = appointment.datetime_when.strftime(DATE_FORMAT_SHORT)
            appointment_date_time = appointment.datetime_when.strftime(DATE_FORMAT_TIME)
            # /-->
        else:
            appointment_date_full = appointment_date_short = appointment_date_time = ""
        return {
            "##A_DATE_FULL##": appointment_date_full,
            "##A_DATE_SHORT##": appointment_date_short,
            "##A_TIME##": appointment_date_time,
            "##A_FLYING_TEAM##": str(appointment.flying_team),
            "##A_STATUS##": appointment.get_status_display(),
            "##A_LOCATION##": appointment.location.name,
            "##A_LOCATION_OR_FLYINGTEAM##": str(appointment.flying_team) or appointment.location.name,
            "##A_WORKER##": str(appointment.worker_assigned),
            '##A_WORKER_PHONE##': worker_phone_number,
            '##A_WORKER_EMAIL##': worker_email_address,
            "##A_ROOM##": str(appointment.room),
            "##A_LENGTH##": str(appointment.length),
            "##A_TYPES##": ", ".join([a.description for a in appointment.appointment_types.all()])
        }

    @staticmethod
    def get_visit_replacements(visit):
        if visit is not None:
            return {
                # TODO: 2to3
                # Was:
                # "##V_DATE_START_FULL##": visit.datetime_begin.strftime(DATETIME_FORMAT).decode(date_format_encoding()),
                # "##V_DATE_START_SHORT##": visit.datetime_begin.strftime(DATE_FORMAT_SHORT).decode(
                #    date_format_encoding()),
                # "##V_DATE_ENDS_FULL##": visit.datetime_end.strftime(DATETIME_FORMAT).decode(date_format_encoding()),
                # "##V_DATE_ENDS_SHORT##": visit.datetime_end.strftime(DATE_FORMAT_SHORT).decode(date_format_encoding()),
                # Is:
                "##V_DATE_START_FULL##": visit.datetime_begin.strftime(DATETIME_FORMAT),
                "##V_DATE_START_SHORT##": visit.datetime_begin.strftime(DATE_FORMAT_SHORT),
                "##V_DATE_ENDS_FULL##": visit.datetime_end.strftime(DATETIME_FORMAT),
                "##V_DATE_ENDS_SHORT##": visit.datetime_end.strftime(DATE_FORMAT_SHORT),
                # /-->
            }
        return {}

    @staticmethod
    def get_subject_replacements(study_subject):
        result = {}
        if study_subject is not None:
            date_born = date_to_str(study_subject.subject.date_born, DATE_FORMAT_SHORT)
            result = {
                "##S_FULL_NAME##": str(study_subject),
                "##S_FIRST_NAME##": study_subject.subject.first_name,
                "##S_LAST_NAME##": study_subject.subject.last_name,
                "##S_ADDRESS##": study_subject.subject.address,
                "##S_CITY##": study_subject.subject.city,
                "##S_COUNTRY##": str(study_subject.subject.country),
                "##S_DIAGNOSIS_YEAR##": str(study_subject.get_custom_field_value('Year of diagnosis')),
                "##S_DATE_ADDED##": date_to_str(study_subject.date_added, DATE_FORMAT_SHORT),
                "##S_DATE_BORN##": date_born,
                "##S_DIAGNOSIS##": str(study_subject.get_custom_field_value('Diagnosis')),
                "##S_EMAIL##": str(study_subject.subject.email),
                "##S_SEX##": study_subject.subject.get_sex_display(),
                "##S_MPOWER_ID##": str(study_subject.get_custom_field_value('MPower ID')),
                "##S_ND_NUMBER##": study_subject.nd_number,
                "##S_PHONE_NUMBER##": str(study_subject.subject.phone_number),
                "##S_PHONE_NUMBER_2##": str(study_subject.subject.phone_number_2),
                "##S_PHONE_NUMBER_3##": str(study_subject.subject.phone_number_3),
                "##S_POST_CODE##": study_subject.subject.postal_code,
                "##S_SCREENING_NUMBER##": study_subject.screening_number,
                "##S_TYPE##": study_subject.get_type_display(),
                '##S_MAIL_LANGUAGE##': str(study_subject.subject.default_written_communication_language),
                '##S_KNOWN_LANGUAGES##': ", ".join([l.name for l in study_subject.subject.languages.all()])
            }
            if study_subject.health_partner is not None:
                result["##S_HEALTH_PARTNER_NAME##"] = str(study_subject.health_partner.name)
                result["##S_HEALTH_PARTNER_ADDRESS##"] = str(study_subject.health_partner.address)
                result["##S_HEALTH_PARTNER_ZIP_CODE##"] = str(study_subject.health_partner.postal_code)
                result["##S_HEALTH_PARTNER_CITY##"] = str(study_subject.health_partner.city)
            else:
                result["##S_HEALTH_PARTNER_NAME##"] = ""
                result["##S_HEALTH_PARTNER_ADDRESS##"] = ""
                result["##S_HEALTH_PARTNER_ZIP_CODE##"] = ""
                result["##S_HEALTH_PARTNER_CITY##"] = ""

            for i in range(1, 6):
                virus_test_field = "##S_VIRUS_{}_RESULT##".format(i)
                virus_test_value = study_subject.get_custom_field_value('Virus {} RT-PCR'.format(i))
                if virus_test_value == "" or virus_test_value is None:
                    virus_test_value = "N/A"

                virus_test_date_field = "##S_VIRUS_{}_SAMPLE_COLLECTION_DATE##".format(i)
                virus_test_date_value = study_subject.get_custom_field_value(
                    'Visit {} RT-PCR collection date'.format(i))

                virus_iga_status_field = "##S_VIRUS_{}_IGA_STATUS##".format(i)
                virus_iga_status_value = study_subject.get_custom_field_value('Visit {} IgA Status'.format(i))
                if virus_iga_status_value is None:
                    virus_iga_status_value = ""

                virus_igg_status_field = "##S_VIRUS_{}_IGG_STATUS##".format(i)
                virus_igg_status_value = study_subject.get_custom_field_value('Visit {} IgG Status'.format(i))
                if virus_igg_status_value is None:
                    virus_igg_status_value = ""

                if virus_test_date_value != "":
                    if virus_iga_status_value == "":
                        virus_iga_status_value = "N/A"
                    if virus_igg_status_value == "":
                        virus_igg_status_value = "N/A"

                result[virus_test_field] = virus_test_value
                result[virus_test_date_field] = virus_test_date_value
                result[virus_iga_status_field] = virus_iga_status_value
                result[virus_igg_status_field] = virus_igg_status_value

        return result

    @staticmethod
    def get_voucher_replacements(voucher):
        if voucher is not None:
            voucher_type = voucher.voucher_type.description
            if voucher.activity_type != '':
                voucher_type += ' (' + voucher.activity_type + ")"
            return {
                "##C_NUMBER##": voucher.number,
                "##C_PATIENT_NAME##": voucher.study_subject.subject.first_name + ' ' + voucher.study_subject.subject.last_name,
                "##C_VOUCHER_TYPE##": voucher_type,
                # TODO: 2to3
                # Was:
                # "##C_ISSUE_DATE_SHORT##": voucher.issue_date.strftime(DATE_FORMAT_SHORT).decode(date_format_encoding()),
                # "##C_EXPIRY_START_SHORT##": voucher.expiry_date.strftime(DATE_FORMAT_SHORT).decode(
                #    date_format_encoding()),
                # Is:
                "##C_ISSUE_DATE_SHORT##": voucher.issue_date.strftime(DATE_FORMAT_SHORT),
                "##C_EXPIRY_START_SHORT##": voucher.expiry_date.strftime(DATE_FORMAT_SHORT),
                # /-->
                "##C_PARTNER_NAME##": voucher.usage_partner.first_name + ' ' + voucher.usage_partner.last_name,
                "##C_PARTNER_ADDRESS##": voucher.usage_partner.address,
                "##C_PARTNER_POSTAL_CODE##": voucher.usage_partner.postal_code,
                "##C_PARTNER_CITY##": voucher.usage_partner.city,
                "##C_PARTNER_COUNTRY##": str(voucher.usage_partner.country),
                "##C_PARTNER_PHONE##": voucher.usage_partner.phone_number,
                "##C_HOURS##": str(voucher.hours),
            }
        return {}
